# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:28 2019

@author: yoelr
"""
from biosteam import System
import biosteam as bst
import thermosteam as tmo
from thermosteam import Stream
from biorefineries.cornstover.process_settings import price, ethanol_density_kggal
from biorefineries.cornstover.chemicals import cornstover_chemicals, chemical_groups
from biorefineries.cornstover.tea import CornstoverTEA
from biorefineries.cornstover import units
import thermosteam.reaction as rxn
import numpy as np

bst.CE = 525.4
System.maxiter = 200
System.molar_tolerance = 1

Ethanol_MW = cornstover_chemicals.Ethanol.MW
Water_MW = cornstover_chemicals.Water.MW

def Ethanol_molfrac(e):
    """Return ethanol mol fraction in a ethanol water mixture"""
    return e/Ethanol_MW / (e/Ethanol_MW + (1-e)/Water_MW)

def find_split(IDs, flow0, flow1):
    flow0 = np.asarray(flow0)
    splits = flow0/(flow0 + np.asarray(flow1))
    thermo = tmo.settings.get_thermo()
    chemicals = thermo.chemicals
    array = np.zeros(chemicals.size)
    for ID, split in zip(IDs, splits):
        if ID in chemical_groups:
            array[chemicals.get_index(chemical_groups[ID])] = split
        else:
            array[chemicals.index(ID)] = split
    return array


# %% Streams

bst.main_flowsheet.set_flowsheet(bst.Flowsheet('cornstover'))
pretreatment_chemical_IDs = ['Acetate', 'AceticAcid', 'Arabinan', 'Ash', 'Cellulase',
                             'Ethanol', 'Extract', 'Furfural', 'Glucan', 'Glucose',
                             'GlucoseOligomer', 'Water', 'H2SO4', 'HMF', 'Lignin',
                             'Mannan', 'NH3', 'Protein', 'SolubleLignin', 'Sucrose',
                             'Xylan', 'Xylose', 'Arabinose', 'XyloseOligomer',
                             'ArabinoseOligomer', 'Mannose', 'MannoseOligomer',
                             'Galactan', 'Galactose', 'GalactoseOligomer']
pretreatment_chemicals = cornstover_chemicals.subgroup(pretreatment_chemical_IDs)
tmo.settings.set_thermo(pretreatment_chemicals)

# tmo.Stream.default_ID_number = 100
    
# feed flow
dry_composition = pretreatment_chemicals.kwarray(
    dict(Glucan=0.3505, Xylan=0.1953, Lignin=0.1576,
         Ash=0.0493, Acetate=0.0181, Protein=0.0310,
         Extract=0.1465, Arabinan=0.0238, Galactan=0.0143,
         Mannan=0.0060, Sucrose=0.0077)
)
moisture_content = pretreatment_chemicals.kwarray(
    dict(Water=0.20)
)
netflow = 104167.0
feedflow = netflow*(dry_composition*0.8 + moisture_content)

cornstover = Stream('cornstover',
                    feedflow,
                    units='kg/hr',
                    price=price['Feedstock'])
warm_process_water = Stream('warm_process_water',
                         T=368.15,
                         P=4.7*101325,
                         Water=140000,
                         units='kg/hr')
rectifier_bottoms_product = Stream('',
                                   T=114+273.15,
                                   P=6.1*101325,
                                   Ethanol=18,
                                   Water=36629,
                                   Furfural=72,
                                   HMF=100,
                                   units='kg/hr')
sulfuric_acid = Stream('sulfuric_acid',
                       P=5.4*101325,
                       T=294.15,
                       Water=139,
                       SulfuricAcid=1842,
                       units='kg/hr',
                       price=price['Sulfuric acid'])
steam = Stream('steam',
               phase='g',
               T=268+273.15,
               P=13*101325,
               Water=24534+3490,
               units='kg/hr')
ammonia = Stream('ammonia',
                 Ammonia=1051,
                 units='kg/hr',
                 phase='l',
                 price=price['Ammonia'])
cellulase = Stream('cellulase',
                   units='kg/hr',
                   price=price['Enzyme'])

# %% Pretreatment system

U101 = units.FeedStockHandling('U101', ins=cornstover)
U101.cost_items['System'].cost = 0

# tmo.Stream.default_ID_number = 200

T201 = units.SulfuricAcidTank('T201', ins=sulfuric_acid)
M201 = units.SulfuricAcidMixer('M201', ins=(rectifier_bottoms_product, T201-0))
M202 = bst.Mixer('M202', ins=(M201-0, warm_process_water, U101-0))
M203 = units.SteamMixer('M203', ins=(M202-0, steam), P=5.5*101325)
R201 = units.PretreatmentReactorSystem('R201', ins=M203-0)
P201 = units.BlowdownDischargePump('P201', ins=R201-1)
T202 = units.OligomerConversionTank('T202', ins=P201-0)
F201 = units.PretreatmentFlash('F201', ins=T202-0, P=101325, Q=0)
M204 = bst.Mixer('M204', ins=(R201-0, F201-0))
H201 = units.WasteVaporCondenser('H201', ins=M204-0, T=99+273.15, V=0)
M210 = units.AmmoniaMixer('M210', ins=(ammonia, warm_process_water))
M205 = bst.Mixer('M205', ins=(F201-1, M210-0))
T203 = units.AmmoniaAdditionTank('T203', ins=M205-0)

# tmo.Stream.default_ID_number = 300

H301 = units.HydrolysateCooler('H301', ins=T203-0, T=48+273.15)
M301 = units.EnzymeHydrolysateMixer('M301', ins=(H301-0, cellulase))

sulfuric_acid_over_feed = sulfuric_acid.mol/cornstover.F_mass
def update_sulfuric_acid_loading():
    # Also plant air
    F_mass_feed = cornstover.F_mass
    plant_air.mol[0] = 0.8 *F_mass_feed
    sulfuric_acid.mol[:] = sulfuric_acid_over_feed * F_mass_feed

hydrolyzate = F201.outs[1]
ammonia_over_hydrolyzate = ammonia.mol/310026.22446428984
def update_ammonia_loading():
    ammonia.mol[:] = ammonia_over_hydrolyzate * hydrolyzate.F_mass

cooled_hydrolyzate = H301.outs[0]
enzyme_over_cellulose = 20/1000 * 20 # (20 g enzyme / cellulose) / (50 g cellulase / 1 L enzyme)
water_cellulase_mass = cellulase.imass['Water', 'Cellulase']
water_cellulase_to_cellulase = np.array([0.95, 0.05])
def update_cellulase_and_nutrient_loading():
    F_mas_cooled_hydrolyzate = cooled_hydrolyzate.F_mass
    cellulose = cornstover.imass['Glucan']
    # Note: An additional 10% is produced for the media glucose/sophorose mixture
    # Humbird (2011) pg. 37 
    water_cellulase_mass[:] = (enzyme_over_cellulose
                               * water_cellulase_to_cellulase
                               * cellulose * 1.1)
    DAP1.mol[:] = DAP1_over_hydrolyzate * F_mas_cooled_hydrolyzate
    DAP2.mol[:] = DAP2_over_hydrolyzate * F_mas_cooled_hydrolyzate
    CSL1.mol[:] = CSL1_over_hydrolyzate * F_mas_cooled_hydrolyzate
    CSL2.mol[:] = CSL2_over_hydrolyzate * F_mas_cooled_hydrolyzate
    
pretreatment_sys = System('pretreatment_sys',
               path=(U101, T201, M201, M202, M203,
                        R201, P201, T202, F201, M204,
                        H201, M210, M205, T203, T203,
                        H301,
                        M301))

# %% Fermentation system

fermentation_chemical_IDs = ['Acetate', 'AceticAcid', 'Arabinan', 'Ash', 'CO2', 'CSL',
                             'Cellobiose', 'DAP', 'Denaturant', 'Enzyme', 'Ethanol',
                             'Extract', 'Furfural', 'Glucan', 'Glucose', 'GlucoseOligomer', 
                             'Glycerol', 'Water', 'H2SO4', 'HMF', 'LacticAcid', 'Lignin',
                             'Mannan', 'NH3', 'O2', 'Protein', 'SolubleLignin',
                             'SuccinicAcid', 'Sucrose', 'Xylan', 'Xylitol', 'Xylose',
                             'XyloseOligomer', 'Z_mobilis', 'Arabinose', 'Mannose',
                             'Galactan', 'Galactose', 'GalactoseOligomer',
                             'ArabinoseOligomer', 'MannoseOligomer']

fermentation_chemicals = cornstover_chemicals.subgroup(fermentation_chemical_IDs)
tmo.settings.set_thermo(fermentation_chemicals)

DAP1 = Stream('DAP1',
                DAP=26,
                units='kg/hr',
                price=price['DAP'])
DAP2 = Stream('DAP2',
                DAP=116,
                units='kg/hr',
                price=price['DAP'])
DAP_storage = units.DAPTank('DAP_storage', ins=Stream('DAP_fresh'), outs='DAP')

S301 = bst.ReversedSplitter('S301', ins=DAP_storage-0, outs=(DAP1, DAP2))
CSL1 = Stream('CSL1',
                CSL=211,
                units='kg/hr',
                price=price['CSL'])
CSL2 = Stream('CSL2',
                CSL=948,
                units='kg/hr',
                price=price['CSL'])
CSL_storage = units.CSLTank('CSL_storage', ins=Stream('CSL_fresh'), outs='CSL')

S302 = bst.ReversedSplitter('S302', ins=CSL_storage-0, outs=(CSL1, CSL2))
denaturant = Stream('denaturant',
                    Octane=230.69,
                    units='kg/hr',
                    price=price['Denaturant'])
stripping_water = Stream('stripping_water',
                          Water=26836,
                          units='kg/hr')
DAP1_over_hydrolyzate = DAP1.mol/451077.22446428984
DAP2_over_hydrolyzate = DAP2.mol/451077.22446428984
CSL1_over_hydrolyzate = CSL1.mol/451077.22446428984
CSL2_over_hydrolyzate = CSL2.mol/451077.22446428984

J1 = upstream=M301-0 - bst.Junction('J1') - Stream()
M302 = bst.Mixer('M302', ins=(J1-0, None))
R301 = units.SaccharificationAndCoFermentation('R301', ins=(M302-0, CSL2, DAP2))
M303 = bst.Mixer('M303', ins=(R301-2, CSL1, DAP1))
R302 = units.SeedTrain('R302', ins=M303-0)
T301 = units.SeedHoldTank('T301', ins=R302-1)
T301-0-1-M302

fermentation_sys = System('fermentation_sys',
               path=(update_cellulase_and_nutrient_loading,
                        J1, M302, R301, M303, R302, T301),
               recycle=M302-0)

# %% Ethanol purification

M304 = bst.Mixer('M304', ins=(R302-0, R301-0))
T302 = units.BeerTank('T302')

# tmo.Stream.default_ID_number = 400

M401 = bst.Mixer('M401', ins=(R301-1, None))
M401-0-T302

D401 = bst.VentScrubber('D401', ins=(stripping_water, M304-0),
                        gas=('CO2', 'NH3', 'O2'))
D401-1-1-M401

# Heat up before beer column
# Exchange heat with stillage
H401 = bst.HXprocess('H401', ins=(T302-0, None),
                    fluid_type='ss', U=1.28)

# Beer column
xbot = Ethanol_molfrac(0.00001)
ytop = Ethanol_molfrac(0.50)
D402 = bst.BinaryDistillation('D402', ins=H401-0, k=1.25,
                       P=101325, y_top=ytop, x_bot=xbot,
                       LHK=('Ethanol', 'Water'))
D402.tray_material = 'Stainless steel 304'
D402.vessel_material = 'Stainless steel 304'
D402.BM = 2.4
D402.boiler.U = 1.85
P401 = bst.Pump('P401', ins=D402-1)
P401-0-1-H401

# Mix ethanol Recycle (Set-up)
M402 = bst.Mixer('M402', ins=(D402-0, None))

ytop = Ethanol_molfrac(0.915)
D403 = bst.BinaryDistillation('D403', ins=M402-0,
                       P=101325, y_top=ytop, x_bot=xbot,
                       k=1.25, LHK=('Ethanol', 'Water'))
D403.tray_material = 'Stainless steel 304'
D403.vessel_material = 'Stainless steel 304'
D403.is_divided = True
D403.boiler.U = 1.85
D403.BM = 2.8
P402 = bst.Pump('P402', ins=D403-1)

JX = P402-0 - bst.Junction("JX") - 0**M201

# Superheat vapor for mol sieve
H402 = bst.HXutility('H402', ins=D403-0, T=115+273.15, V=1)

# Molecular sieve
U401 = bst.MolecularSieve('U401', ins=H402-0,
                         split=(2165.14/13356.04, 1280.06/1383.85),
                         order=('Ethanol', 'Water'))

U401-0-1-M402

ethanol_recycle_sys = System('ethanol_recycle_sys',
                             path=(M402, D403, H402, U401),
                             recycle=M402-0)

# Condense ethanol product
H403 = bst.HXutility('H403', ins=U401-1, V=0, T=350.)

# IDnum_400 = tmo.Stream.default_ID_number

# tmo.Stream.default_ID_number = 700
T701 = bst.StorageTank('T701', ins=H403-0, tau=7*24,
                       vessel_type='Floating roof',
                       vessel_material='Carbon steel')
P701 = bst.Pump('P701', ins=T701-0)

# Storage for gasoline
T702 = bst.StorageTank('T702', ins=denaturant, tau=7*24,
                       vessel_type='Floating roof',
                       vessel_material='Carbon steel')
P702 = bst.Pump('P702', ins=T702-0)

# Mix in denaturant
ethanol = Stream('ethanol', price=price['Ethanol'])
M701 = bst.MixTank('M701', ins=(P702-0, P701-0), outs=ethanol)
M701.line = 'Mixer'
M701.tau = 0.05

def adjust_denaturant():
    denaturant.imol['Octane'] = 0.022*P701.outs[0].F_mass/114.232

P401.BM = P402.BM = P701.BM = P702.BM = 3.1
T701.BM = T702.BM = 1.7

vent_stream = M304-0
stripping_water_over_vent = stripping_water.mol / 21202.490455845436
def update_stripping_water():
    stripping_water.mol[:] = stripping_water_over_vent * vent_stream.F_mass

puresys = System('purification',
                 path=(M304,
                          update_stripping_water,
                          D401, 
                          M401, T302,
                          H401, D402,
                          H401, P401, H401,
                          ethanol_recycle_sys,
                          P402, H403, T701, P701,
                          adjust_denaturant,
                          T702, P702, M701, JX))

# %% Lignin Separation
tmo.settings.set_thermo(cornstover_chemicals)

recycled_water = tmo.Stream(Water=1,
                            T=47+273.15,
                            P=3.9*101325,
                            units='kg/hr')

splits = [('Glucose', 19, 502),
          ('Xylose', 40, 1022),
          ('OtherSugars', 81, 2175),
          ('SugarOligomers', 60, 1552),
          ('OrganicSolubleSolids', 612, 15808),
          ('InorganicSolubleSolids', 97, 2513),
          ('Furfurals', 19, 513),
          ('OtherOrganics', 52, 1348),
          ('Glucan', 1230, 25),
          ('Xylan', 415, 8),
          ('OtherStructuralCarbohydrates', 94, 2),
          ('Lignin', 12226, 250),
          ('Protein', 3376, 69),
          ('CellMass', 925, 19),
          ('OtherInsolubleSolids', 4489, 92)]

# tmo.Stream.default_ID_number = IDnum_400

S401 = units.PressureFilter('S401', ins=('', recycled_water),
                            moisture_content=0.35,
                            split=find_split(*zip(*splits)))
J2 = H401-1 - bst.Junction('J2') - 0**S401

# %% Waste water treatment

Hvap_water = cornstover_chemicals.Water.Hvap(298.15, 101325)
combustion = cornstover_chemicals.get_combustion_reactions()
def growth(reactant):
    f = cornstover_chemicals.WWTsludge.MW / getattr(cornstover_chemicals, reactant).MW
    return rxn.Reaction(f"{f}{reactant} -> WWTsludge", reactant, 1.)
    
organic_groups = ['OtherSugars', 'SugarOligomers', 'OrganicSolubleSolids',
                  'Furfurals', 'OtherOrganics', 'Protein', 'CellMass']
organics = list(sum([chemical_groups[i] for i in organic_groups],
                    ('Ethanol', 'AceticAcid', 'Xylose', 'Glucose')))
organics.remove('WWTsludge')

P_sludge = 0.05/0.91/cornstover_chemicals.WWTsludge.MW
MW = np.array([cornstover_chemicals.CH4.MW, cornstover_chemicals.CO2.MW])
mass = np.array([0.51, 0.49])*MW
mass /= mass.sum()
mass *= 0.86/(0.91)
P_ch4, P_co2 = mass/MW
def anaerobic_rxn(reactant):
    MW = getattr(cornstover_chemicals, reactant).MW
    return rxn.Reaction(f"{1/MW}{reactant} -> {P_ch4}CH4 + {P_co2}CO2 + {P_sludge}WWTsludge",
                        reactant, 0.91)

# TODO: Revise this with Jeremy
anaerobic_digestion = rxn.ParallelReaction([anaerobic_rxn(i) for i in organics] + 
                                           [rxn.Reaction(f"H2SO4 -> H2S + 2O2", 'H2SO4', 1.)])


# Note, nitogenous species included here, but most of it removed in R601 digester
# TODO: Add ammonium to reaction, make sure it can be a liquid, possibly add Henry's constant
aerobic_digestion = rxn.ParallelReaction([i*0.74 + 0.22*growth(i.reactant)
                                          for i in combustion
                                          if (i.reactant in organics)])
aerobic_digestion.X[:] = 0.96

splits = [('Ethanol', 1, 15),
          ('Water', 27158, 356069),
          ('Glucose', 3, 42),
          ('Xylose', 7, 85),
          ('OtherSugars', 13, 175),
          ('SugarOligomers', 10, 130),
          ('OrganicSolubleSolids', 182, 2387),
          ('InorganicSolubleSolids', 8, 110),
          ('Ammonia', 48, 633),
          ('AceticAcid', 0, 5),
          ('Furfurals', 5, 70),
          ('OtherOrganics', 9, 113),
          ('Cellulose', 19, 6),
          ('Xylan', 6, 2),
          ('OtherStructuralCarbohydrates', 1, 0),
          ('Lignin', 186, 64),
          ('Protein', 51, 18),
          ('CellMass', 813, 280),
          ('OtherInsolubleSolids', 68, 23)]

# tmo.Stream.default_ID_number = 600

well_water = Stream('well_water', Water=1, T=15+273.15)
M601 = bst.Mixer('M601', ins=(S401-1, '', '', ''))
J3 = H201-0 - bst.Junction('J3') - 1**M601

WWTC = units.WasteWaterSystemCost('WWTC', ins=M601-0)
R601 = units.AnaerobicDigestion('R601', ins=(WWTC-0, well_water),
                                 reactions=anaerobic_digestion,
                                 sludge_split=find_split(*zip(*splits)))

air = Stream('air_lagoon', O2=51061, N2=168162, phase='g', units='kg/hr')
caustic = Stream('WWT_caustic', Water=2252, NaOH=2252,
                 units='kg/hr', price=price['Caustic']*0.5)
# polymer = Stream('WWT polymer') # Empty in humbird report :-/

M602 = bst.Mixer('M602', ins=(R601-1, None))

caustic_over_waste = caustic.mol / 2544300.6261793654
air_over_waste = air.mol / 2544300.6261793654
waste = M602-0
def update_aerobic_input_streams():
    F_mass_waste = waste.F_mass
    caustic.mol[:] = F_mass_waste * caustic_over_waste
    air.mol[:] = F_mass_waste * air_over_waste

R602 = units.AerobicDigestion('R602', ins=(waste, air, caustic),
                              outs=('evaporated_water', ''),
                              reactions=aerobic_digestion)

splits = [('Ethanol', 0, 1),
          ('Water', 381300, 2241169),
          ('Glucose', 0, 2),
          ('Xylose', 1, 3),
          ('OtherSugars', 1, 7),
          ('SugarOligomers', 1, 6),
          ('OrganicSolubleSolids', 79, 466),
          ('InorganicSolubleSolids', 4828, 28378),
          ('Ammonia', 3, 16),
          ('Furfurals', 0, 3),
          ('OtherOrganics', 1, 7),
          ('CarbonDioxide', 6, 38),
          ('O2', 3, 17),
          ('N2', 5, 32),
          ('Cellulose', 0, 194),
          ('Xylan', 0, 65),
          ('OtherStructuralCarbohydrates', 0, 15),
          ('Lignin', 0, 1925),
          ('Protein', 0, 90),
          ('CellMass', 0, 19778),
          ('OtherInsolubleSolids', 0, 707)]

S601 = bst.Splitter('S601', ins=R602-1, split=find_split(*zip(*splits)))

S602 = bst.Splitter('S602', ins=S601-1, split=0.96)

M603 = bst.Mixer('M603', ins=(S602-0, None))
M603-0-1-M602

M604 = bst.Mixer('M604', ins=(R601-2, S602-1))

centrifuge_species = ('Water', 'Glucose', 'Xylose', 'OtherSugars',
                      'SugarOligomers', 'OrganicSolubleSolids',
                      'InorganicSolubleSolids', 'Ammonia', 'Furfurals', 
                      'OtherOrganics', 'CO2', 'COxSOxNOxH2S', 'Cellulose',
                      'Xylan', 'OtherStructuralCarbohydrates', 'Lignin',
                      'Protein', 'CellMass', 'OtherInsolubleSolids')
S623_flow = np.array([7708, 0, 0, 1, 1, 13, 75, 3, 0, 1, 1, 2, 25, 8, 2, 250, 52, 1523, 92])
S616_flow = np.array([109098, 3, 6, 13, 9, 187, 1068, 46, 5, 8, 14, 31, 1, 0, 0, 13, 3, 80, 5])

S603 = bst.Splitter('S603', ins=M604-0, outs=('', 'sludge'),
                  split=find_split(centrifuge_species, S616_flow, S623_flow))
S603-0-1-M603

S604 = bst.Splitter('S604', ins=S601-0, outs=('treated_water', 'waste_brine'),
                  split={'Water': 0.987})

aerobic_digestion_sys = System('aerobic_digestion_sys',
                               path=(M602, R602, S601, S602, M604, S603, M603),
                               recycle=M602-0)

# %% Facilities

# tmo.Stream.default_ID_number = 500

M501 = bst.Mixer('M501', ins=(S603-1, S401-0))
BT = bst.facilities.BoilerTurbogenerator('BT', ins=(M501-0, R601-0), 
                                         turbogenerator_efficiency=0.85)
BT.outs[-1].T = 373.15

# tmo.Stream.default_ID_number = 700

CWP = bst.facilities.ChilledWaterPackage('CWP')
CT = bst.facilities.CoolingTower('CT')
CT.outs[1].T = 273.15 + 28
water_thermo = tmo.Thermo(tmo.Chemicals(['Water']))

process_water_streams = (caustic,
                         stripping_water,
                         warm_process_water,
                         steam, BT-1, CT-1)
        
makeup_water = Stream('makeup_water', thermo=water_thermo, price=price['Makeup water'])

PWC = bst.facilities.ProcessWaterCenter('PWC',
                                        (S604-0, makeup_water),
                                        (),
                                        (BT-1, CT-1),
                                        process_water_streams)
J4 = BT.outs[-1] - bst.Junction('J4') - 2**M601
J5 = CT.outs[1] - bst.Junction('J5') - 3**M601

Substance = tmo.Chemical.blank('Substance')
Substance.at_state(phase='l')
Substance.default()
substance_thermo = tmo.Thermo(tmo.Chemicals([Substance]))
ash = Stream('ash', thermo=substance_thermo,
             price=price['Ash disposal'])
lime = Stream('lime', thermo=substance_thermo,
             price=price['FGD lime'])
boilerchems = Stream('boiler_chemicals', thermo=substance_thermo,
                     price=price['Boiler chems'])
emission = BT.outs[0]
def update_lime_boilerchems_and_ash():
    emission_ash = emission.imol['Ash']
    lime.imol['Substance'] = lime_flow = emission_ash * 0.21
    ash.imol['Substance'] = (emission_ash + lime_flow) * 1.18 # Include lime and other stuff
    boilerchems.imol['Substance'] = 0.24620/865 * lime_flow

CIP = Stream('CIP', thermo=substance_thermo, flow=(126,))
CIP_package = units.CIPpackage('CIP_package', ins=CIP, thermo=substance_thermo)

plant_air = Stream('plant_air', flow=(83333,), thermo=substance_thermo)

ADP = bst.facilities.AirDistributionPackage('ADP', ins=plant_air, thermo=substance_thermo)

FT = units.FireWaterTank('FT',
                         ins=Stream('fire_water', flow=(8343,), thermo=substance_thermo),
                         thermo=substance_thermo)

# %% Complete system

cornstover_sys = System('cornstover_sys',
                        path=(pretreatment_sys, fermentation_sys, puresys, J2, S401,
                                 J3, J4, J5, M601, WWTC, R601,
                                 update_aerobic_input_streams,
                                 aerobic_digestion_sys, S604),
                        facilities=(M501, CWP, BT, CT,
                                    PWC, ADP, update_lime_boilerchems_and_ash,
                                    CIP_package, S301, S302, DAP_storage,
                                    CSL_storage, FT))
cornstover_sys.products.add(ash)
baghouse_bags = Stream(ID='Baghouse_bags', thermo=substance_thermo, flow=(1,), price=11.1)
cornstover_sys.feeds.add(lime)
cornstover_sys.feeds.add(boilerchems)
cornstover_sys.feeds.add(baghouse_bags)
aerobic_digestion_sys.converge_method = 'Fixed point'
for i in range(3): cornstover_sys.simulate()
ethanol_tea = CornstoverTEA(
        system=cornstover_sys, 
        IRR=0.10, 
        duration=(2007, 2037),
        depreciation='MACRS7', 
        income_tax=0.35,
        operating_days=350.4,
        lang_factor=None, 
        construction_schedule=(0.08, 0.60, 0.32),
        startup_months=3, 
        startup_FOCfrac=1,
        startup_salesfrac=0.5,
        startup_VOCfrac=0.75,
        WC_over_FCI=0.05,
        finance_interest=0.08,
        finance_years=10,
        finance_fraction=0.4,
        OSBL_units=(WWTC, CWP, CT, PWC, ADP), # BT not included
        warehouse=0.04, 
        site_development=0.09, 
        additional_piping=0.045,
        proratable_costs=0.10,
        field_expenses=0.10,
        construction=0.20,
        contingency=0.10,
        other_indirect_costs=0.10, 
        labor_cost=2.5e6,
        labor_burden=0.90,
        property_insurance=0.007, 
        maintenance=0.03)
ethanol_tea.units.remove(BT)
ethanol_tea.units.remove(U101)
Area700 = bst.TEA.like(System('Area700', (BT,)),
                       ethanol_tea)
Area700.labor_cost = 0
Area700.depreciation = 'MACRS20'
Area700.OSBL_units = (BT,)
cornstover_tea = bst.CombinedTEA([ethanol_tea, Area700], IRR=0.10)
cornstover_sys._TEA = cornstover_tea
ethanol.price = cornstover_tea.solve_price(ethanol, ethanol_tea)
ethanol.price = cornstover_tea.solve_price(ethanol, ethanol_tea)
ethanol_price_USDgal = ethanol.price * ethanol_density_kggal

# %% Areas

Area100 = bst.TEA.like(System(None, (U101,)), ethanol_tea)
Area200 = bst.TEA.like(System(None,
                              (T201, M201, R201, P201,
                               T202, F201, H201, M210, T203)),
                       ethanol_tea)
Area300 = bst.TEA.like(System(None, 
                              (H301, M301, R301,
                               R302, T301, T302)),
                       ethanol_tea)
Area400 = bst.TEA.like(System(None, 
                              (D401, H401, D402, P401,
                               M402, D403, P402, H402,
                               U401, H403, M701, S401)),
                       ethanol_tea)
Area500 = bst.TEA.like(System(None, (WWTC,)),
                       ethanol_tea)
Area600 = bst.TEA.like(System(None,
                              (T701, T702, P701, P702, M701, FT,
                               CSL_storage, DAP_storage)),
                       ethanol_tea) 
Area800 = bst.TEA.like(System(None, (CWP, CT, PWC, ADP, CIP_package)),
                       ethanol_tea)
areas = (Area100, Area200, Area300, Area400,
         Area500, Area600, Area700, Area800)
numbered_areas = tuple(enumerate(areas, 1))
installation_costs = {i: j.installation_cost/1e6
                      for i,j in numbered_areas}
utility_costs = {i: j.utility_cost/1e6
                 for i,j in numbered_areas}

def get_utility(units, ID, attr):
    out = 0
    for i in units:
        for j in i.heat_utilities:
            if j.ID == ID:
                out += getattr(j, attr)
    return out

get_rate = lambda units: sum([i.power_utility.rate
                              for i in units])/1e3

get_ecost = lambda units: sum([i.power_utility.cost
                               for i in units])*24*350.4/1e6

cooling_water_uses = {i: get_utility(j.units, 'Cooling water', 'duty')/1e6/4.184
                      for i,j in numbered_areas}
electricity_uses = {i: get_rate(j.units)/41 for i,j in numbered_areas}
electricity_costs = {i: get_ecost(j.units) for i,j in numbered_areas}




# -*- coding: utf-8 -*-
"""
Created on Tue Aug 13 21:44:23 2019

@author: yoelr
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
from biosteam.utils import colors
from biosteam.evaluation.evaluation_tools import plot_single_points, plot_horizontal_line, \
                                                 plot_montecarlo, plot_vertical_line

data = pd.read_excel('Monte Carlo cornstover.xlsx', header=[0, 1])

# %% Plot MESP
# plt.figure()

# posistions_MESP = (0,)
# MESP_data = data[('Biorefinery', 'Minimum ethanol selling price')]
# bx_MESP = plot_montecarlo(MESP_data,
#                           colors.blue_tint.RGBn,
#                           colors.blue_shade.RGBn, 
#                           posistions_MESP, transpose=False)
# dot_MESP = plot_single_points(posistions_MESP, [2.15], s=125, color=colors.red_dark.RGBn)
# plt.ylabel('MESP ($\mathrm{USD} \cdot \mathrm{gal}^{-1}$)')
# plt.ylim(1.5, 2.50)
# bx_patch = Patch(facecolor=colors.blue_tint.RGBn, edgecolor=colors.blue_shade.RGBn)
# plt.legend([bx_patch, dot_MESP], ['BioSTEAM', 'benchmark'])
# plt.xticks([], [])
                                                 
# %% Setup of subplots

# light_color = colors.blue_tint.RGBn
# dark_color = colors.blue_shade.RGBn
# dot_color = colors.purple_shade.RGBn
nums = tuple(range(1, 9))

fig, axes = plt.subplots(ncols=1, nrows=2, constrained_layout=True, gridspec_kw=dict(height_ratios=[1, 4]))
magnitude_ax, metric_over_benchmark_ax = axes
plt.sca(metric_over_benchmark_ax)

# %% Plot MESP and ethanol sales

positions_other = (16, 17, 18)
other_index = [('Biorefinery', 'Steam demand [kg/hr]'),
               ('Biorefinery', 'Ethanol production [kg/hr]'),
               ('Biorefinery', 'Minimum ethanol selling price [USD/gal]')]
other_data = np.array(data[other_index])
other_data[:, 0] *= 100./234784.
other_data[:, 1] *= 100./22273.
other_data[:, 2] *= 100./2.15
bx_other = plot_montecarlo(other_data,
                           colors.blue_tint.RGBn,
                           colors.blue_shade.RGBn, positions_other, transpose=False)

# %% Plot electricity

plot_vertical_line(15.5, color=colors.grey_tint.RGBn)
positions_electricity = tuple(range(9))
areas = [f'Area {i}00' for i in nums]
units = 'MW'
positions = np.arange(0, 9)
electricity_cols = [(i, 'Electricity [MW]') for i in areas]
humbird_electricity = 41 * np.array((0.02, 0.14, 0.06, 0.05,
                                     0.18, 0.003, 0.03, 0.08,
                                     0.44))
electricity_data = data[electricity_cols]  #/ humbird_electricity
electricity_data[('Biorefinery', 'Excess electricity [MW]')] = data[('Biorefinery', 'Excess electricity [MW]')]
electricity_data_humbird_normalized = electricity_data * (100/humbird_electricity)
bx_electricity = plot_montecarlo(electricity_data_humbird_normalized,
                                 colors.orange_tint.RGBn,
                                 colors.orange_shade.RGBn,
                                 transpose=True, positions=positions_electricity)

# plot_vertical_line(7.5, color=colors.orange_tint.shade(15).RGBn, ls='-.')
# plot_vertical_line(9.5, color=colors.orange_tint.shade(15).RGBn, ls='-.')

# %% Plot installation cost

units = '10^6 USD'
plot_vertical_line(8.5, color=colors.grey_tint.RGBn)
positions_installation = tuple(range(9, 16))
installation_cols = [(i, 'Installation cost [10^6 USD]') for i in areas[1:]]
humbird_installation = np.array([24.2, 32.9, 31.2, 22.3, 49.4, 5, 66, 6.9])
installation_data = data[installation_cols]
# installation_data[('Biorefinery', 'Installation cost')] = installation_data.sum(1)
installation_data_humbird_normalized = installation_data * (100/humbird_installation[1:]/1e6)
bx_installation = plot_montecarlo(installation_data_humbird_normalized,
                                  colors.purple_tint.RGBn, colors.purple_shade.RGBn,
                                  transpose=True, positions=positions_installation)
plot_horizontal_line(100, ls='--')
u_lb = 0; y_ub = 250
plt.ylim(0, 250)
yticks = np.arange(0, 251, 50)
plt.yticks(yticks)
y_text = 0.885*y_ub
y_letter = 0.875 * y_ub
plt.text(4, y_text, "Electricity demand", color=colors.orange_shade.RGBn,
          horizontalalignment='center', fontsize=12, fontweight='bold')
plt.text(-0.25, y_letter, "C", color=colors.neutral_shade.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
plt.text(12, y_text, "Installation cost", color=colors.purple_shade.RGBn,
          horizontalalignment='center', fontsize=12, fontweight='bold')
plt.text(8.75, y_letter, "D", color=colors.neutral_shade.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
plt.text(15.75, y_letter, "E", color=colors.neutral_shade.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
# plt.text(9.25, y_letter, "D", color=colors.neutral_shade.RGBn,
#          horizontalalignment='center', fontsize=14, fontweight='bold')

# plt.text(16, y_text, "Ethanol\nsales", color=colors.red_shade.RGBn,
#           horizontalalignment='center', fontsize=12, fontweight='bold')
# plt.text(16.0, y_text, "MESP", color=colors.blue_shade.RGBn,
#           horizontalalignment='center', fontsize=12, fontweight='bold')

plt.xlim(-0.5, 18.5)
plt.ylabel("Metric over benchmark [%]")
area_marks = [i.replace(' ', '\n') for i in areas]
xmarks = area_marks + ['Excess'] + area_marks[1:] + ['Steam\ndemand', '   EtOH\n    prod.', '  MESP']
xticks = positions_electricity + positions_installation + positions_other
plt.xticks(xticks, xmarks)
metric_over_benchmark_ax.set_zorder(1e6)

plt.sca(magnitude_ax)
plt.fill_between([-0.5, 15.5], 0, 1, color=colors.neutral_tint.tint(85).RGBn)

electricity_areas = humbird_electricity # electricity_data.median()
electricity_areas /= max(electricity_areas)
plt.bar(positions_electricity, electricity_areas, 0.5,
        align='center', label="Electricity demand",
        color=colors.orange.tint(30).shade(15).RGBn,
        edgecolor=colors.orange_shade.RGBn)
plot_vertical_line(8.5, color=colors.grey_tint.RGBn)

installation_areas = humbird_installation[1:]# installation_data.median()
installation_areas /= max(installation_areas)
plt.bar(positions_installation, installation_areas, 0.5,
        align='center', label="Installation cost",
        color=colors.purple.tint(30).shade(15).RGBn,
        edgecolor=colors.purple_shade.RGBn)

plot_vertical_line(15.5, color=colors.grey_tint.RGBn)
# plot_vertical_line(15.5, color='k', lw=0.8)
# plt.bar(positions_MESP, [1], 0.5,
#         align='center', label="MESP",
#         color=colors.blue.tint(30).shade(15).RGBn,
#         edgecolor=colors.blue_shade.RGBn)

plot_vertical_line(-0.5, color='k')
plt.hlines([1], [-0.5], [15.5], color='k')
magnitude_ax.spines['top'].set_visible(False)
magnitude_ax.spines['right'].set_visible(False)
magnitude_ax.tick_params(axis="x", direction="in", length=0)
magnitude_ax.set_zorder(2)
metric_over_benchmark_ax.set_zorder(1)
plt.yticks([], [])
plt.xticks(xticks[:-3], [])
plt.ylim(0, 1)
plt.xlim(-0.5, 18.5)
plt.text(4, 0.65, "Benchmark magnitude", color=colors.neutral.shade(35).RGBn,
         horizontalalignment='center', fontsize=12, fontweight='bold')
plt.text(-0.25, 0.60, "A", color=colors.neutral_shade.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
plt.text(8.75, 0.60, "B", color=colors.neutral_shade.RGBn,
         horizontalalignment='left', fontsize=16, fontweight='bold')
plt.subplots_adjust(hspace=.0)

metric_over_benchmark_ax.tick_params(axis='x', direction="inout", length=4)
for ax in axes:
    ax.tick_params(axis='y', right=False, direction="inout", length=4)
ax2 = metric_over_benchmark_ax.twinx()
plt.sca(ax2)
plt.ylim(0, 250)
plt.yticks(yticks, [])
ax2.zorder = 1000
ax2.tick_params(direction="in")
    
xlabels = metric_over_benchmark_ax.get_xticklabels()    
# for xtick in xlabels[-3:]:
#     xtick.set_rotation(90)


# plot_vertical_line(15.5, color=colors.purple.tint(20).shade(10).RGBn, ls='-.')

# leg1 = ax.legend([bx_economic['boxes'][0]], ['MESP'], loc="upper left")
# leg2 = ax.legend([bx_electricity['boxes'][0]], ['Electricity'], loc="upper center")
# leg3 = ax.legend([bx_installation['boxes'][0]], ['Installation'], loc="upper right")
# ax.add_artist(leg2)
# ax.add_artist(leg1)

# light_box = Patch(color=colors.neutral_tint.RGBn)
# line = Line2D([0], [0], color=colors.neutral_shade.RGBn)
# dark_box = Patch(color=colors.neutral_shade.RGBn)
# plt.legend([(light_box, line), dark_box], ["Metric over benchmark (%)", "Relative benchmark magnintude"])

# leg1 = ax.legend([bx_economic['boxes'][0]], ['MESP'], loc="upper left")


# -*- coding: utf-8 -*-
"""
Created on Thu Jun 27 23:12:04 2019

@author: yoelr
"""
import thermosteam as tmo
from biorefineries.lipidcane.chemicals import lipidcane_chemicals
from thermosteam import functional as fn
import pandas as pd

__all__ = ('cornstover_chemicals', 'get_grouped_chemicals', 'chemical_groups')

# %%  Constants

# Common structural carbohydrates properties

# Assume heat capacity of lignin, cellulose, and hemicellulose
# and all components at 350 K are about the same [2,2].
Cp_cellulosic = 1.364

# Assume density is similar for most solids
rho = 1540 # kg/m3

cal2joule = 4.184

# %% Initialize Chemicals object and define functions

chems = cornstover_chemicals = tmo.Chemicals([])

# Chemicals to define
chemical_IDs = [
        'Water', 'Ethanol', 'Glucose', 'Galactose',
        'Mannose', 'Xylose', 'Arabinose', 'Cellobiose',
        'Sucrose', 'GlucoseOligomer', 'GalactoseOligomer',
        'MannoseOligomer', 'XyloseOligomer', 'ArabinoseOligomer',
        'Extract','SolubleLignin','HMF', 'Furfural', 'AceticAcid',
        'LacticAcid', 'Xylitol', 'Glycerol', 'SuccinicAcid',
        'NH3', 'H2SO4', 'NH4SO4', 'AmmoniumAcetate', 'DAP',
        'HNO3', 'NaNO3', 'NaOH', 'CellulaseNutrients',
        'Denaturant', 'Oil', 'Cellulose', 'Galactan', 'Mannan', 'Glucan',
        'Xylan', 'Arabinan', 'Lignin', 'Acetate', 'Protein',
        'Ash', 'Enzyme', 'DenaturedEnzyme', 'Z_mobilis', 'T_reesei',
        'Biomass', 'Tar', 'CaO', 'CaSO4', 'Graphite', 'N2', 'O2', 'CO2',
        'CH4', 'H2S', 'SO2', 'NO', 'CO', 'AmmoniumSulfate', 'NO2', 'CSL',
        'WWTsludge', 'Cellulase'
]

def append_single_phase_chemical(ID, search_ID=None, **data):
    chemical = tmo.Chemical(ID, search_ID=search_ID, **data)
    try: chemical.at_state(phase=chemical.phase_ref)
    except: pass
    chemical.default()    
    chems.append(chemical)

def extend_single_phase_chemicals(IDs):
    for ID in IDs: append_single_phase_chemical(ID)

def append_new_single_phase_chemical(ID, source=None, **data):
    chemical = tmo.Chemical.blank(ID, **data)
    if source: 
        default_phase_ref = source.phase_ref
        chemical.copy_models_from(source)
    else:
        default_phase_ref = 'l'
    if not chemical.phase_ref:
        chemical.phase_ref = default_phase_ref
    chemical.at_state(chemical.phase_ref)
    chemical.default()
    chems.append(chemical)

def append_chemical_copy(ID, chemical):
    new_chemical = chemical.copy(ID)
    chems.append(new_chemical)

def set_Cp(single_phase_chemical, Cp):
    chem = single_phase_chemical
    chem.Cn.add_model(Cp * chem.MW, top_priority=True)

def set_rho(single_phase_chemical, rho):
    V = fn.rho_to_V(rho, single_phase_chemical.MW)
    single_phase_chemical.V.add_model(V, top_priority=True)

# %% Define species

# TODO: Add heats of combustion

# As is in data bank
chems.extend(
    tmo.Chemicals(['Water', 'Ethanol', 'AceticAcid',
                   'Furfural', 'Glycerol', 'H2SO4',
                   'LacticAcid', 'SuccinicAcid', 'P4O10'])
)
chems.P4O10.at_state('s')
chems.H2SO4.at_state('l')
append_single_phase_chemical('HNO3', 'NitricAcid')
append_single_phase_chemical('Denaturant', 'Octane')
append_single_phase_chemical('DAP', 'Diammonium Phosphate')
append_single_phase_chemical('AmmoniumAcetate')
append_single_phase_chemical('AmmoniumSulfate')
append_single_phase_chemical('NaNO3', 'SodiumNitrate')
append_single_phase_chemical('Oil', 'Oleic acid')
append_single_phase_chemical('HMF')

# Will remain in the vapor phase
extend_single_phase_chemicals(['N2', 'NH3', 'O2', 'CH4', 'H2S', 'SO2'])
append_single_phase_chemical('CO2')

# Analagous vapors
append_new_single_phase_chemical('NO2', chems.N2,
                                 formula='NO2',
                                 Hf=7925*cal2joule)
append_new_single_phase_chemical('NO', chems.N2,
                                 formula='NO',
                                 Hf=82.05)
append_single_phase_chemical('CO', 'Carbon monoxide', Hf=-110.522)

# Will remain as  solid
extend_single_phase_chemicals(['Glucose', 'Xylose', 'Sucrose'])
append_single_phase_chemical('CaSO4')

subgroup = chems['Glucose', 'Xylose', 'Sucrose', 'CaSO4', 'AmmoniumSulfate']
for chemical in subgroup: set_Cp(chemical, Cp_cellulosic)

# Analagous sugars
append_chemical_copy('Mannose', chems.Glucose)
append_chemical_copy('Galactose', chems.Glucose)
append_chemical_copy('Arabinose', chems.Xylose)

# Other analogues
append_chemical_copy('CellulaseNutrients', chems.Glucose)
append_chemical_copy('Extract', chems.Glucose)
append_chemical_copy('Acetate', chems.AceticAcid)
append_chemical_copy('Tar', chems.Xylose)
chems.Acetate.Hf = -103373

# Chemicals taken from previous study
chems.append(lipidcane_chemicals.CaO)
chems.append(lipidcane_chemicals.Ash)
chems.append(lipidcane_chemicals.NaOH)
append_new_single_phase_chemical('Lignin',
                                 formula='C8H8O3',
                                 Hf=-108248*cal2joule, MW=152.15)
set_rho(chems.Lignin, 1540)
set_Cp(chems.Lignin, Cp_cellulosic)
append_chemical_copy('SolubleLignin', chems.Lignin)

# Create structural carbohydrates
append_chemical_copy('GlucoseOligomer', chems.Glucose)
set_Cp(chems.GlucoseOligomer, Cp_cellulosic)
chems.GlucoseOligomer.formula = "C6H10O5"
chems.GlucoseOligomer.Hf = -233200*cal2joule

append_chemical_copy('GalactoseOligomer', chems.GlucoseOligomer)
append_chemical_copy('MannoseOligomer', chems.GlucoseOligomer)
append_chemical_copy('XyloseOligomer', chems.Xylose)
set_Cp(chems.XyloseOligomer, Cp_cellulosic)
chems.XyloseOligomer.formula = "C5H8O4"
chems.XyloseOligomer.Hf = -182100*cal2joule

append_chemical_copy('ArabinoseOligomer', chems.XyloseOligomer)

# Other
append_new_single_phase_chemical('Z_mobilis', formula="CH1.8O0.5N0.2",
                                 Hf=-31169.39*cal2joule)
append_new_single_phase_chemical('T_reesei', formula="CH1.645O0.445N0.205S0.005",
                                 Hf=-23200.01*cal2joule)
append_new_single_phase_chemical('Biomass', formula="CH1.64O0.39N0.23S0.0035",
                                 Hf=-23200.01*cal2joule)
append_new_single_phase_chemical('Cellulose', formula="C6H10O5", # Glucose monomer minus water
                                 Hf=-233200.06*cal2joule)
append_new_single_phase_chemical('Protein', formula="CH1.57O0.31N0.29S0.007",
                                 Hf=-17618*cal2joule)
append_new_single_phase_chemical('Enzyme', formula="CH1.59O0.42N0.24S0.01",
                                 Hf=-17618*cal2joule)
append_new_single_phase_chemical('Glucan', formula='C6H10O5',
                                 Hf=-233200*cal2joule)
append_new_single_phase_chemical('Xylan', formula="C5H8O4",
                                 Hf=-182100*cal2joule)
append_new_single_phase_chemical('Xylitol', formula="C5H12O5",
                                 Hf=-243145*cal2joule)
append_new_single_phase_chemical('Cellobiose', formula="C12H22O11",
                                 Hf=-480900*cal2joule)
append_new_single_phase_chemical('CSL', MW=1,
                                 Hf=(chems.Protein.Hf/4
                                     + chems.Water.Hf/2
                                     + chems.LacticAcid.Hf/4))
append_chemical_copy('DenaturedEnzyme', chems.Enzyme)
append_chemical_copy('Arabinan', chems.Xylan)
append_chemical_copy('Mannan',   chems.Glucan)
append_chemical_copy('Galactan', chems.Glucan)

# %% TODO: Maybe remove this

# For waste water
append_chemical_copy('WWTsludge', chems.Biomass)
append_chemical_copy('Cellulase', chems.Enzyme)

# %% Grouped chemicals

chemical_groups = dict(
    OtherSugars = ('Arabinose',
                   'Mannose',
                   'Galactose',
                   'Cellobiose',
                   'Sucrose'),
    SugarOligomers = ('GlucoseOligomer',
                      'XyloseOligomer',
                      'GalactoseOligomer',
                      'ArabinoseOligomer',
                      'MannoseOligomer'),
    OrganicSolubleSolids = ('AmmoniumAcetate',
                            'SolubleLignin',
                            'Extract', 
                            'LacticAcid', 
                            'Cellulase'),
    InorganicSolubleSolids = ('AmmoniumSulfate',
                              'DAP',
                              'NaOH',
                              'HNO3',
                              'NaNO3'),
    Furfurals = ('Furfural',
                 'HMF'),
    OtherOrganics = ('Glycerol',
                     'Denaturant',
                     'Oil',
                     'SuccinicAcid',
                     'Xylitol'),
    COxSOxNOxH2S = ('NO',
                    'NO2',
                    'SO2',
                    'CO',
                    'H2S'),
    Protein = ('Protein',
               'Enzyme',
               'DenaturedEnzyme'),
    CellMass = ('WWTsludge',
                'Z_mobilis',
                'T_reesei'),
    OtherInsolubleSolids = ('Tar',
                            'Ash',
                            'Lime'),
    OtherStructuralCarbohydrates = ('Arabinan', 
                                    'Mannan', 
                                    'Galactan')
)

tmo.settings.set_thermo(chems)
thermo = tmo.settings.get_thermo()
chems.set_synonym('CaO', 'Lime')
chems.set_synonym('Water', 'H2O')
chems.set_synonym('H2SO4', 'SulfuricAcid')
chems.set_synonym('NH3', 'Ammonia')
chems.set_synonym('AmmoniumSulfate', 'NH4SO4')
chems.set_synonym('Denaturant', 'Octane')
chems.set_synonym('CO2', 'CarbonDioxide')

def get_grouped_chemicals(stream, units='kmol/hr'):
    new_stream = tmo.Stream(thermo=thermo)
    new_stream.set_flow(stream.mol, units, stream.chemicals.IDs)
    data = {group: new_stream.get_flow(units, IDs).sum() for group, IDs in chemical_groups.items()}
    return pd.Series(data)

# Fix sugar properties


# cellulosic = ('Glucan', 'Xylan', 'GlucoseOligomer','XyloseOligomer',
#               'Acetate', 'SolubleLignin')




# -*- coding: utf-8 -*-
"""
Created on Tue Feb  4 06:42:02 2020

@author: yoelr
"""
from thermosteam import functional as fn
import thermosteam as tmo

__all__ = ('lipidcane_chemicals', 'pretreatment_chemicals',
           'ethanol_chemicals', 'biodiesel_chemicals')


# %% Define common chemicals
Biodiesel = tmo.Chemical('Biodiesel',
                         search_ID='Methyl oleate')
lipidcane_chemicals = tmo.Chemicals(
    ['Water', 'Methanol', 'Ethanol', 'Glycerol',
     'Glucose', 'Sucrose', 'H3PO4', 'P4O10', 'CO2',
     'Octane', 'O2', Biodiesel])

(Water, Methanol, Ethanol,
 Glycerol, Glucose, Sucrose,
 H3PO4, P4O10, CO2, Octane, O2, Biodiesel) = lipidcane_chemicals

O2.at_state(phase='g')
CO2.at_state(phase='g')
H3PO4.at_state(phase='s')
P4O10.at_state(phase='s')
Glucose.at_state(phase='s')
Sucrose.at_state(phase='s')

# %% Define new chemicals

def create_new_chemical(ID, phase='s', **constants):
    solid = tmo.Chemical.blank(ID, phase=phase, **constants)
    lipidcane_chemicals.append(solid)
    return solid

Ash = create_new_chemical('Ash', MW=1.)
Cellulose = create_new_chemical('Cellulose',
                                formula="C6H10O5", # Glucose monomer minus water
                                MW=162.14,
                                Hf=-975708.8)
Hemicellulose = create_new_chemical('Hemicellulose',
                                    formula="C5H8O5", # Xylose monomer minus water
                                    MW=132.12,
                                    Hf=-761906.4)
Flocculant = create_new_chemical('Flocculant',
                                 MW=1.)
Lignin = create_new_chemical('Lignin',
                             formula='C8H8O3', # Vainillin
                             MW=152.15,
                             Hf=-452909.632)
Solids = create_new_chemical('Solids', MW=1.)
DryYeast = create_new_chemical('DryYeast', MW=1., CAS='Yeast')
CaO = create_new_chemical('CaO', MW=56.0774)
HCl = create_new_chemical('HCl', MW=36.46094)
NaOH = create_new_chemical('NaOH', MW=39.997109)
NaOCH3 = create_new_chemical('NaOCH3', MW=54.023689)

Lipid = create_new_chemical(
    'Lipid',
    phase = 'l',
    MW = 885.432,
    formula = 'C57H104O6',
    Hf = -2193.7e3
)
Lipid.Dortmund.set_group_counts_by_name({'CH3':3, 'CH2':41, 'CH':1, 'CH=CH':3, 'CH2COO':3})

# %% Fill missing properties

# Assume properties are similar for trioleate and tripalmitin
Tripalmitin = tmo.Chemical('Tripalmitin').at_state(phase='l', copy=True)
Lipid.copy_models_from(Tripalmitin, ['V', 'sigma', 'kappa', 'Cn'])

# Assume a constant volume for lipid
lipid_molar_volume = fn.rho_to_V(rho=900, MW=Lipid.MW)
Lipid.V.add_model(lipid_molar_volume)

# Insolubles occupy a significant volume
insoluble_solids = (Ash, Cellulose, Hemicellulose, Sucrose,
                    Flocculant, Lignin, Solids, DryYeast, P4O10)

# Solubles don't occupy much volume
soluble_solids = (CaO, HCl, NaOH, H3PO4, Glucose) 

for chemical in insoluble_solids:
    V = fn.rho_to_V(rho=1540, MW=chemical.MW)
    chemical.V.add_model(V, top_priority=True)

for chemical in soluble_solids:
    V = fn.rho_to_V(rho=1e5, MW=chemical.MW)
    chemical.V.add_model(V, top_priority=True)

# Assume sodium methoxide has some of the same properities as methanol
LiquidMethanol = Methanol.at_state(phase='l', copy=True)
NaOCH3.copy_models_from(LiquidMethanol, ['V', 'sigma', 'kappa', 'Cn'])

# Add constant models for molar heat capacity of solids
Ash.Cn.add_model(0.09 * 4.184 * Ash.MW) 
CaO.Cn.add_model(1.02388 * CaO.MW) 
Cellulose.Cn.add_model(1.364 * Cellulose.MW) 
Hemicellulose.Cn.add_model(1.364 * Hemicellulose.MW)
Flocculant.Cn.add_model(4.184 * Flocculant.MW)
Lignin.Cn.add_model(1.364 * Lignin.MW)
Solids.Cn.add_model(1.100 * Solids.MW)

for chemical in lipidcane_chemicals: chemical.default()

lipidcane_chemicals.compile()
lipidcane_chemicals.set_synonym('Water', 'H2O')

pretreatment_chemicals = lipidcane_chemicals.subgroup(
    ['Lipid', 'Ash', 'Cellulose',
     'Flocculant', 'Hemicellulose',
     'Lignin', 'Solids', 'CaO',
     'Water', 'Ethanol',
     'Glucose', 'Sucrose',
     'DryYeast', 'H3PO4', 'O2', 'CO2', 'P4O10']
)
ethanol_chemicals = lipidcane_chemicals.subgroup(
    ['Water', 'CO2', 'Ethanol',
     'Glucose', 'Sucrose', 'DryYeast',
     'H3PO4', 'Octane']
)
biodiesel_chemicals = lipidcane_chemicals.subgroup(
    ['Lipid', 'Biodiesel', 'Methanol', 'Glycerol', 'Water',
     'HCl', 'NaOH', 'NaOCH3']
)


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 23:03:19 2019

This module defines the composition_balance function, which performs mass energy balance the given oil composition of lipid cane to determine its composition.

@author: yoelr
"""
import thermosteam as tmo
from thermosteam import Stream
from biorefineries.lipidcane.chemicals import pretreatment_chemicals
from biorefineries.lipidcane.system import lipid_cane
from numpy import array

__all__ = ('set_lipid_fraction',)

tmo.settings.set_thermo(pretreatment_chemicals)
getattr_ = getattr
carbs_IDs = ('Glucose', 'Sucrose')
fiber_IDs = ('Lignin', 'Cellulose', 'Hemicellulose')
lipid_IDs = ('Lipid',)
water_IDs = ('Water',)

lc = lipid_cane
carbs   = Stream('Carbs')
fiber   = Stream('Fiber')
lipid   = Stream('Lipid')
carbs.imol[carbs_IDs] = lc.imol[carbs_IDs]
fiber.imol[fiber_IDs] = lc.imol[fiber_IDs]
lipid.imol[lipid_IDs] = lc.imol[lipid_IDs]
streams = (carbs, fiber, lipid)

# Mass property arrays
carbs_mass = lc.imass[carbs_IDs]
fiber_mass = lc.imass[fiber_IDs]
lipid_mass = lc.imass[lipid_IDs]

# Net weight
carbs_massnet = carbs_mass.sum()
fiber_massnet = fiber_mass.sum()
lipid_massnet = lipid_mass.sum()

# Heats of combustion per kg
LHV_carbs_kg = carbs.LHV/carbs_massnet
LHV_lipid_kg = lipid.LHV/lipid_massnet

# Composition
carbs_massfrac = array(carbs_mass/carbs_massnet)
fiber_massfrac = array(fiber_mass/fiber_massnet)

water_mass = lc.imass['Water']
solids_mass = lc.imass['Solids']
ash_mass = lc.imass['Ash']

def set_lipid_fraction(lipid_fraction):
    """Adjust composition of lipid cane to achieve desired oil fraction (dry weight)."""
    z_mass_lipid = lipid_fraction
    F_mass = lc.F_mass
    z_dry = 0.3
    z_mass_new_lipid = z_mass_lipid * z_dry 
    z_mass_new_carbs = 0.149 - z_mass_new_lipid * (LHV_lipid_kg / LHV_carbs_kg)
    new_fiber_massfrac = z_dry - z_mass_new_carbs - z_mass_new_lipid - 0.006 - 0.015
    
    water_mass.value = F_mass*0.7
    ash_mass.value = F_mass*0.006
    solids_mass.value = F_mass*0.015
    lipid_mass[:] = z_mass_new_lipid * F_mass
    carbs_mass[:] = carbs_massfrac * z_mass_new_carbs * F_mass
    fiber_mass[:] = fiber_massfrac * new_fiber_massfrac * F_mass
    if any(lc.mol < 0):
        raise ValueError(f'lipid cane oil composition of {z_mass_lipid*100:.0f}% dry weight is infeasible')


### Old attempt to include energy balance. It does not work because fiber is more energy dense than sugar (to my surprise) ###

# LHV_0 = lc.LHV
# def LHV_error(carbs_massnet, dryweight_no_oil):
#     """Return the difference between the original heat of combustion and the heat of combustion at the new carbohydrate composition"""
#     carbs_mass[:] = carbs_massfrac * carbs_massnet
#     fiber_mass[:] = fiber_massfrac * (dryweight_no_oil - carbs_massnet)
#     return LHV_0 - lc.LHV

# def composition_balance(oilfrac):
#     """Adjust composition of lipid cane to achieve desired oil fraction (dry weight)."""
#     global carbs_massnet
#     arg = oilfrac
#     oil_mass[:] = mass_oil = oilfrac * dryweight
#     dryweight_no_oil = dryweight - mass_oil
#     carbs_massnet = newton(LHV_error, carbs_massnet, args=(dryweight_no_oil,))
#     if any(lc.mol < 0):
#         raise ValueError(f'Lipid cane oil composition of {arg*100:.0f}% dry weight is infeasible')
#     return carbs_massnet
     
# composition_balance(0.1)
# check = lambda: oil_mass.sum()/dryweight
    


# -*- coding: utf-8 -*-
"""
Created on Sun Sep  1 04:37:16 2019

@author: yoelr
"""
import numpy as np
import biorefineries.lipidcane as lc
import biorefineries.sugarcane as sc

def run_uncertainty(N_spearman_samples = 5000,
                    N_coordinate_samples = 1000,
                    N_coordinates = 40):
    
    model_lc = lc.model.lipidcane_model
    model_sc = sc.model.sugarcane_model
    model_lc_lf = lc.model.lipidcane_model_with_lipidfraction_parameter
    np.random.seed(1234)
    rule = 'L'
    
    # Monte Carlo across lipid fraction    
    coordinate = np.linspace(0.11, 0.01, N_coordinates)
    samples = model_lc.sample(N_coordinate_samples, rule)
    model_lc.load_samples(samples)
    
    model_lc.evaluate_across_coordinate('Lipid fraction',
          lc.utils.set_lipid_fraction, coordinate,
          xlfile='Monte Carlo across lipid fraction.xlsx')
        
    # Sugar cane Monte Carlo    
    samples = model_sc.sample(N_coordinate_samples, rule)
    model_sc.load_samples(samples)
    model_sc.evaluate()
    model_sc.table.to_excel('Monte Carlo sugarcane.xlsx')

    if N_spearman_samples:
        # Spearman's correlation    
        samples = model_lc_lf.sample(N_spearman_samples, rule)
        model_lc_lf.load_samples(samples)
        model_lc_lf.evaluate()
        IRR_metric = model_lc_lf.metrics[0]
        spearman = model_lc_lf.spearman(metrics=(IRR_metric,))
        spearman.to_excel("Spearman correlation lipidcane.xlsx")

def run_without_uncertainty(N_coordinates = 40):
    model_lc = lc.model.lipidcane_model
    model_sc = sc.model.sugarcane_model
    coordinate = np.linspace(0.15, 0.01, N_coordinates)
    
    # Lipid cane
    sample = model_lc.get_baseline_sample()
    samples = np.expand_dims(sample, axis=0)
    model_lc.load_samples(samples)
    model_lc.evaluate_across_coordinate('Lipid fraction',
          lc.utils.set_lipid_fraction, coordinate,
          xlfile='Monte Carlo across lipid fraction.xlsx',
          notify=False)
    
    # Sugar cane
    sample = model_sc.get_baseline_sample()
    samples = np.expand_dims(sample, axis=0)
    model_sc.load_samples(samples)
    model_sc.evaluate()
    model_sc.table.to_excel('Monte Carlo sugarcane.xlsx')




